﻿using System;
using System.Collections.Generic;
using HeuristicLab.Core;
using HeuristicLab.Hive.JobBase;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Hive.Experiment.Jobs {
  [StorableClass]
  public class OptimizerJob : IJob {
    [Storable]
    protected Guid internalId;
    public Guid InternalId {
      get { return internalId; }
    }
    
    [Storable]
    protected IOptimizer optimizer;
    public IOptimizer Optimizer {
      get { return optimizer; }
      set {
        if (optimizer != null && value != optimizer) {
          DeregisterEvents();
        }
        optimizer = value;
        if (optimizer != null) {
          RegisterEvents();
        }
      }
    }

    [Storable]
    protected double progress = 0.0;

    [Storable]
    protected ILog log;
    public ILog Log {
      get { return log; }
    }

    [Storable]
    protected ICollection<IJob> childJobs;
    public IEnumerable<IJob> ChildJobs {
      get { return childJobs; }
    }

    [Storable]
    protected bool computeInParallel;
    public bool ComputeInParallel {
      get { return computeInParallel; }
      set { computeInParallel = value; }
    }

    [Storable]
    protected int coresNeeded;
    public int CoresNeeded {
      get { return coresNeeded; }
      set { coresNeeded = value; }
    }

    [Storable]
    protected int memoryNeeded;
    public int MemoryNeeded {
      get { return memoryNeeded; }
      set { memoryNeeded = value; }
    }

    public OptimizerJob() {
      this.internalId = Guid.NewGuid();
      this.log = new Log();
      this.childJobs = new List<IJob>();
      this.ComputeInParallel = false;
      this.CoresNeeded = 1;
      this.MemoryNeeded = 0;
    }

    public OptimizerJob(IOptimizer optimizer)
      : this() {
      this.optimizer = optimizer;
    }

    [StorableHook(HookType.AfterDeserialization)]
    protected virtual void AfterDeserialization() {
      RegisterEvents();
    }

    #region IJob Members
    
    public double Progress {
      get { return optimizer.ExecutionState == ExecutionState.Stopped ? 1.0 : this.progress; }
    }

    public virtual ExecutionState ExecutionState {
      get { return optimizer.ExecutionState; }
    }

    public virtual void Run() {
      throw new NotSupportedException();
    }

    public virtual void Prepare() {
      optimizer.Prepare();
    }

    public virtual void Start() {
      optimizer.Start();
    }

    public virtual void Stop() {
      optimizer.Stop();
    }

    public virtual void Resume(IEnumerable<IJob> childJobs) {
      throw new NotSupportedException();
    }

    public event EventHandler JobStopped;
    protected virtual void OnJobStopped() {
      EventHandler handler = JobStopped;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler JobFailed;
    protected virtual void OnJobFailed(Common.EventArgs<Exception> e) {
      EventHandler handler = JobFailed;
      if (handler != null) handler(this, e);
    }

    public event EventHandler<Common.EventArgs<IJob>> NewChildJob;
    protected virtual void OnNewChildJob(IJob job) {
      EventHandler<Common.EventArgs<IJob>> handler = NewChildJob;
      if (handler != null) handler(this, new Common.EventArgs<IJob>(job));
    }

    public event EventHandler WaitForChildJobs;
    protected virtual void OnWaitForChildJobs() {
      EventHandler handler = WaitForChildJobs;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    #endregion

    #region Events
    protected virtual void RegisterEvents() {
      optimizer.Stopped += new EventHandler(optimizer_Stopped);
      optimizer.ExceptionOccurred += new EventHandler<Common.EventArgs<Exception>>(optimizer_ExceptionOccurred);
      optimizer.DescriptionChanged += this.DescriptionChanged;
      optimizer.FilenameChanged += this.FilenameChanged;
      optimizer.ItemImageChanged += this.ItemImageChanged;
      optimizer.NameChanged += this.NameChanged;
      optimizer.NameChanging += this.NameChanging;
      optimizer.ToStringChanged += this.ToStringChanged;
    }

    protected virtual void DeregisterEvents() {
      optimizer.Stopped -= new EventHandler(optimizer_Stopped);
      optimizer.ExceptionOccurred -= new EventHandler<Common.EventArgs<Exception>>(optimizer_ExceptionOccurred);
      optimizer.DescriptionChanged -= this.DescriptionChanged;
      optimizer.FilenameChanged -= this.FilenameChanged;
      optimizer.ItemImageChanged -= this.ItemImageChanged;
      optimizer.NameChanged -= this.NameChanged;
      optimizer.NameChanging -= this.NameChanging;
      optimizer.ToStringChanged -= this.ToStringChanged;
    }

    protected virtual void optimizer_ExceptionOccurred(object sender, Common.EventArgs<Exception> e) {
      if (optimizer.ExecutionState != ExecutionState.Stopped) this.progress = 0.0;
      else this.progress = 1.0;
      OnJobFailed(e);
    }

    protected virtual void optimizer_Stopped(object sender, EventArgs e) {
      if (optimizer.ExecutionState != ExecutionState.Stopped) this.progress = 0.0;
      else this.progress = 1.0;
      OnJobStopped();
    }
    #endregion


    #region INamedItem Members

    public bool CanChangeDescription {
      get { return optimizer.CanChangeDescription; }
    }

    public bool CanChangeName {
      get { return optimizer.CanChangeName; }
    }

    public string Description {
      get { return optimizer.Description; }
      set { optimizer.Description = value; }
    }
    public event EventHandler DescriptionChanged;

    public string Name {
      get { return optimizer.Name; }
      set { optimizer.Name = value; }
    }
    public event EventHandler NameChanged;
    public event EventHandler<Common.CancelEventArgs<string>> NameChanging;

    #endregion

    #region IItem Members

    public string ItemDescription {
      get { return optimizer.ItemDescription; }
    }

    public System.Drawing.Image ItemImage {
      get { return optimizer.ItemImage; }
    }

    public event EventHandler ItemImageChanged;
    public string ItemName {
      get { return optimizer.ItemName; }
    }

    public Version ItemVersion {
      get { return optimizer.ItemVersion; }
    }
    public event EventHandler ToStringChanged;

    #endregion

    #region IStorableContent Members

    public string Filename {
      get;
      set;
    }
    public event EventHandler FilenameChanged;

    #endregion

    #region IDeepCloneable Members

    public Common.IDeepCloneable Clone(Common.Cloner cloner) {
      throw new NotImplementedException();
    }

    #endregion

    #region ICloneable Members

    public object Clone() {
      throw new NotImplementedException();
    }

    #endregion

    /// <summary>
    /// Gets the string representation of the current instance in the format: <c>Name: [null|Value]</c>.
    /// </summary>
    /// <returns>The current instance as a string.</returns>
    public override string ToString() {
      return Name;
    }
    
    public virtual bool IsParallelizable {
      get { return false; }
    }
  }
}
